@@ -1,10 +1,10 @@
 # sqlalchemy/log.py
-# Copyright (C) 2006-2020 the SQLAlchemy authors and contributors
+# Copyright (C) 2006-2022 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 # Includes alterations by Vinay Sajip vinay_sajip@yahoo.co.uk
 #
 # This module is part of SQLAlchemy and is released under
-# the MIT License: http://www.opensource.org/licenses/mit-license.php
+# the MIT License: https://www.opensource.org/licenses/mit-license.php
 
 """Logging control and utilities.
 
@@ -21,6 +21,17 @@ instance only.
 import logging
 import sys
 
+from .util import py311
+from .util import py38
+
+if py38:
+    STACKLEVEL = True
+    # needed as of py3.11.0b1
+    # #8019
+    STACKLEVEL_OFFSET = 2 if py311 else 1
+else:
+    STACKLEVEL = False
+    STACKLEVEL_OFFSET = 0
 
 # set initial level to WARN.  This so that
 # log statements don't occur in the absence of explicit
@@ -41,8 +52,15 @@ def _add_default_handler(logger):
 _logged_classes = set()
 
 
+def _qual_logger_name_for_cls(cls):
+    return (
+        getattr(cls, "_sqla_logger_namespace", None)
+        or cls.__module__ + "." + cls.__name__
+    )
+
+
 def class_logger(cls):
-    logger = logging.getLogger(cls.__module__ + "." + cls.__name__)
+    logger = logging.getLogger(_qual_logger_name_for_cls(cls))
     cls._should_log_debug = lambda self: logger.isEnabledFor(logging.DEBUG)
     cls._should_log_info = lambda self: logger.isEnabledFor(logging.INFO)
     cls.logger = logger
@@ -153,6 +171,11 @@ class InstanceLogger(object):
             selected_level = self.logger.getEffectiveLevel()
 
         if level >= selected_level:
+            if STACKLEVEL:
+                kwargs["stacklevel"] = (
+                    kwargs.get("stacklevel", 1) + STACKLEVEL_OFFSET
+                )
+
             self.logger._log(level, msg, args, **kwargs)
 
     def isEnabledFor(self, level):
@@ -175,16 +198,12 @@ def instance_logger(instance, echoflag=None):
     """create a logger for an instance that implements :class:`.Identified`."""
 
     if instance.logging_name:
-        name = "%s.%s.%s" % (
-            instance.__class__.__module__,
-            instance.__class__.__name__,
+        name = "%s.%s" % (
+            _qual_logger_name_for_cls(instance.__class__),
             instance.logging_name,
         )
     else:
-        name = "%s.%s" % (
-            instance.__class__.__module__,
-            instance.__class__.__name__,
-        )
+        name = _qual_logger_name_for_cls(instance.__class__)
 
     instance._echo = echoflag
 
